/*
    (C) Gustav Hartvigsson, 2013.
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** @file */

#ifndef __H_BASE_OBJECT__
#define __H_BASE_OBJECT__

/* ---------------------------------------------------
 * This is just a test of creating a small typesystem.
 * It will include refrence counting.
 * ---------------------------------------------------
 */


typedef struct _SBaseObjectClass SBaseObjectClass;


typedef struct _SBaseObjectInstance SBaseObjectInstance;

/**
 * The class holds all the "virtual" functions, also knows  as methods
 * that are to be used with the object.
 * 
 */
struct _SBaseObjectClass {
  // SBaseObjectInstance * (* initize)(SBaseObjectInstance *);
  void (*deinitize)(SBaseObjectInstance *);
  int (*ref)(SBaseObjectInstance *);
  int (*unref)(SBaseObjectInstance *);
  int (*get_refcount)(SBaseObjectInstance *);
  char * (*to_string)(SBaseObjectInstance *);
};

/**
 * The instance holds the data that is associated with the object.
 * it also holds a pointer to the class of the object.
 * 
 */
struct _SBaseObjectInstance {
  SBaseObjectClass * base_class;
  unsigned int refcount;
};


/* -----------------
 * Helper functions...
 * -----------------
 */
/*
 * This function is used to set the method to initialize a new instance of an
 * object.
 *
 * This has no use, rely...(?)
 */
// void s_base_object_set_init_method (SBaseObjectInstance * self, SBaseObjectInstance * (* method)(* SBaseObjectInstance));

/**
 * This function is used to set the method to deinitialize an object.
 * 
 * set it to a method that deinitialize your object.
 */
void s_base_object_set_deinit_method (SBaseObjectInstance * self, void (* method)(SBaseObjectInstance *));

/**
 * This function is used to set the ref method.
 * 
 * DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING.
 */
void s_base_object_set_ref_method (SBaseObjectInstance * self, int (* method)(SBaseObjectInstance *));

/**
 * This function is used to set the unref method.
 * 
 * DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING.
 */
void s_base_object_set_unref_method (SBaseObjectInstance * self, int (* method)(SBaseObjectInstance *));

/**!
 * This function is used to set the get_refcount method.
 * 
 * DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING.
 */
void s_base_object_set_get_refcount_method (SBaseObjectInstance * self, int (* method)(SBaseObjectInstance *));

/**!
 * This function is used to set the to_string method.
 */
void s_base_object_set_to_string_method (SBaseObjectInstance * self, char * (* method)(SBaseObjectInstance *));


/*
 * concrete functions are defined in the C file.
 */

/* ----------------------
 * Base object functions.
 * ----------------------
 */

/** @brief
 * Initializes an SBaseObjectInstance
 * 
 * @param self The object to initialize.
 * 
 * This function initializes an instance of the SBaseObjectInstance, it also sets
 * the methods to be used with the object and sets the reference count to one.
 */
void s_base_object_initize (SBaseObjectInstance * self);

/** @brief
 * This function creates a new base object.
 * 
 * 
 * 
 * @return a new SBaseObjectInstance
 */
SBaseObjectInstance * s_base_object_new ();

/**
 * This function desensitizes/frees an object even if it is still referenced.
 * This is usually a bad idea, use s_base_object_unref () instead.
 */
void s_base_object_free (SBaseObjectInstance * self);

/**
 * This function gets the class (which hold the object methods).
 */
void * s_base_object_get_class (SBaseObjectInstance * self);

/**
 * This function sets the instance class of an object.
 */
void s_base_object_set_class (SBaseObjectInstance * self, SBaseObjectClass * klass);

/**
 * This function is used to decrese the reference count of an object.
 * When an object reaches zero, it will deinitialize the object using the
 * objects deinitialize method.
 * 
 * It returns the current (after change) reference count.
 */
int s_base_object_unref (SBaseObjectInstance * self);

/**
 * This function is used to increase the reference count of an object.
 *
 * Returns the current (after change) reference count.
 */
int s_base_object_ref (SBaseObjectInstance * self);

/**
 * This function returns the current reference count without changing it.
 */
int s_base_object_get_refcount (SBaseObjectInstance * self);

/**
 * This function returns a textual (string) that represents the object.
 * The method can be set using s_base_object_set_to_string_method.
 *
 * Note: The string that is returned must be freed.
 */
char * s_base_object_to_string (SBaseObjectInstance * self);


#endif
